'use strict';

const express = require('express');
const bodyParser = require('body-parser');
const logger = require('./logger');
//const appmanager = require('./appmanager');
const http = require('http');
const https = require('https');
const async = require('async');
const fs = require("fs");
const request = require('request');
const path = require('path');
const Joi = require('joi');
const uuid = require('uuid-random');
const os = require('os');
const url = require('url');

const apipolcard = require('./restapi-polcard');
const apitools = require('./restapi-tools');
const apiasyncreqs = require('./restapi-asyncreqs');

const yaml_config = require('node-yaml-config');
logger.logger.info('***************************************** STARTING *****************************************');
logger.logger.info('Loading napi, make sure all DLL files are in place ( ldd build/Release/native.node | grep \'not found\' )');
const nativeTest = require('./build/Release/native.node');
logger.logger.info('napi: DONE');

//###################################################
//                  Config
//###################################################
logger.logger.info('Loading config, make sure config.yml has valid structure');
var buildVersion = null;
var schemaVersion = null;
var config = yaml_config.load(__dirname + '/config.yml');
logger.logger.info('config: DONE');

//###################################################
//                  Methods
//###################################################

const app = express();

app.use(bodyParser.json({ limit: '80mb' }));

app.use(function (req, res, next) {
    res.header("Access-Control-Allow-Origin", "*");
    res.header("Access-Control-Allow-Headers", "Origin, X-Requested-With, Content-Type, Accept");
    next();
});

var api_definition = {
    "/v1/license/check": {  //bacward compatible with version 1.3
        mapto: "/license/v1/check",
        deprecated: true
    },
    "/license/v1/check": {
        method: 'get',
        grouptag: 'license',
        func: (req, res) => { processLicenseCheck(req, res); },
        description: 'Pass ?json=true/false to change the output format',
        schema_resp: {
            "ok": "bool: true/false",
            "text": "string",
            "validTo": "string",
            "licenseTo": "string",
            "expired": "bool: true/false"
        }
    },
    "/v1/license/upgrade": {
        mapto: "/license/v1/upgrade",
        deprecated: true
    },
    "/license/v1/upgrade": {
        method: 'post',
        grouptag: 'license',
        func: (req, res) => { processLicenseUpgrade(req, res); },
        schema_resp: {
            "ok": "bool: true/false",
            "text": "string",
            "validTo": "string",
            "licenseTo": "string",
            "expired": "bool: true/false"
        }
    },
    "/v1/ingenico_status": {
        mapto: "/eservice/v1/ingenico_status",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_status": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoStatus(req, res); },
        schema_resp: {
            "ok": "bool: true/false", "watermark" : "string", "error": "string", "terminal": {
                "returnCode": "number: (0,1,2,7) - 0 accepted, 1 refused, 2 No connection, 7 interrupted by the user",
                "transactionStatus": "number: (0x00 - 0x81) " +
                    "0x00 - Idle, " +
	                "0x01 - WaitCard, " +
                    "0x02 - WaitPIN, " +
                    "0x03 - WaitEMVApp, " +
                    "0x04 - WaitHost, " +
                    "0x05 - WaitSign, " +
                    "0x06 - WaitTrEnd, " +
                    "0x07 - WaitNoCard, " +
                    "0x08 - WaitBusy, " +
                    "0x09 - InProgress, " +
                    "0x0A - WaitCopy, " +
                    "0x0B - WaitAuthCode, " +
                    "0x0C - WaitAction, " +
                    "0x0D - BatchCompleted, " +
                    "0x0E - DCCCurrency, " +
                    "0x0F - CashBackAmount, " +
                    "0x10 - Used internaly, " +
                    "0x11 - Used internaly, " +
                    "0x12 - Transaction accepted, " +
                    "0x13 - Used internaly, " +
                    "0x14 - Waiting for amount, " +
                    "0x15 - Waiting for selection, " +
                    "0x16 - Asking ECR to print data, " +
                    "0x17 - Presenting ZenCard offer, " +
                    "0x18 - Waiting for fid, " +
                    "0x19 - Asking ECR to receive data, " +
                    "0x20 - Asking ECR to provide data, " +
                    "0x21 - Zencard continue, " +
                    "0x80 - Application in error state, " +
                    "0x81 - Reconciliation needed ",
                "sequenceNb": ""
            }
        },
    },
    "/polcard/v1/ingenico_status": {
        method: 'get',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processPolcardTransactionStatus(req, res); }
    },
    "/polcard/v1/ingenico_statuses": {
        method: 'get',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processPolcardTransactionStatuses(req, res); }
    },
    "/polcard/v1/ingenico_display": {
        method: 'get',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processPolcardDisplayInfo(req, res); }
    },
    "/v1/ingenico_general_info": {
        mapto: "/eservice/v1/ingenico_general_info",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_general_info": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoGeneralInfo(req, res); },
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string",
            "terminal": {
                "returnCode": "number: (0,1,2,7) - 0 accepted, 1 refused, 2 No connection, 7 interrupted by the user",
                "terminalInfo": "string - comma separated data"
            }
        }
    },
    "/v1/ingenico_auto_transaction_process": {
        mapto: "/eservice/v1/ingenico_auto_transaction_process",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_auto_transaction_process": {
        method: 'post',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionAuto(req, res); },
        schema: Joi.object().keys({
            type: Joi.string().required().valid('purchase', 'preauthorization', 'completion', 'preauthorization_cancel', 'refund', 'reversal', 'giftcard_activation', 'giftcard_balance', 'giftcard_topup', 'reconciliation', 'connection_test', 'tms'),
            amount: Joi.number().required(),
            title: Joi.string(),//.regex(/^\s*\w+(?:[^\w,]+\w+)*[^,\w]*$/),
            /*
            ^ - start of string
            \s* - 0 or more whitespaces (or, use [^,\w]* to match 0 or more chars other than comma and word chars)
            \w+ - 1 or more word chars (letters, digits or _, if you do not want _, replace with [^\W_])
            (?:[^\w,]+\w+)* - zero or more repetitions of
            [^\w,]+ - 1 or more chars other than comma and word chars
            \w+ - 1 or more word chars
            [^,\w]* - 0 or more chars other than comma and word chars
            $ - end of string.
             */
            sequenceNb: Joi.string(),
            callback: Joi.object().keys({
                uri: Joi.string().required(),
                method: Joi.string().required().valid('POST', 'GET', 'PUT')
            }),
            posnetserver: Joi.object(),
            arguments: Joi.object().keys({
                transactionFlags: Joi.array().items(Joi.string().valid('TokenGeneration' /*0x01*/, 'ReferenceNumberGeneration'/*0x02*/, 'TransitTransactionIndicator'/*0x04*/)),     //DF01
                blikCode: Joi.number(),             //DF02 – BLIK code
                additionalPrintout: Joi.string(),   //DF03 – Additional printout data
                hints: Joi.string(),                //DF04 – "Hints"
                rfu: Joi.string(),                  //DF05 – RFU
                variableSymbol: Joi.string(),       //DF06 – Variable Symbol
                specificSymbol: Joi.string(),       //DF07 – Specific Symbol
                saleItemList: Joi.array().items(Joi.object().keys({   //7F01 – Sale Item List
                    itemId: Joi.string(),               //DF01 - Item ID 
                    productCode: Joi.string(),          //DF02 – Product Code 
                    amount: Joi.number(),               //DF03 - Amount(Total cost of item)
                    unitMeasure: Joi.string(),          //DF04 – Unit Measure
                    unitPrice: Joi.number(),            //DF05 – Unit Price(Price of single unit of item) 
                    quantity: Joi.number(),             //DF06 - Quantity 
                    taxCode: Joi.number()               //DF07 – Tax Code
                }))
            })
        }),
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string",
            "uuid": "string"
        },
        description: 'Use posnetserver structure when integration with PosnetServer is enabled (config.posnetserver.active=true). Sample JSON object: {"lines": [{ "na": "Towar 1", "il": 1.0, "vtp": "23,00", "pr": 2350 },{ "na": "Towar 2", "il": 1.0, "vtp": "23,00", "pr": 1150 }], "summary": { "to": 3500 } }. PosnetServer documentation can be found here: https://blog.bigdotsoftware.pl/posnet-server-wydruk-paragonu/'
    },
    "/v1/ingenico_auto_transaction_cancel/:uuid": {
        mapto: "/eservice/v1/ingenico_auto_transaction_cancel/:uuid",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_auto_transaction_cancel/:uuid": {
        method: 'post',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionCancel(req, res); },
        description: 'Method markes request as cancelled and releases COM port. Transaction still have to be manually cancelled from terminal\'s keyboard',
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string",
            "uuid": "string"
        }
    },
    "/v1/ingenico_auto_transaction_status/:uuid": {
        mapto: "/eservice/v1/ingenico_auto_transaction_status/:uuid",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_auto_transaction_status/:uuid": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionStatus(req, res); },
        schema_resp: {
            "watermark": "string",
            "ts": "long",
            "state": "string: success, error, cancelled, processing",
            "req": "object",
            "result": "object"
        }
    },
    "/v1/ingenico_auto_transaction_statuses": {
        mapto: "/eservice/v1/ingenico_auto_transaction_statuses",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_auto_transaction_statuses": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionStatuses(req, res); },
        schema_resp: {
            "watermark": "string",
            "{uuid}": {
                "ts": "long",
                "state": "string: success, error, cancelled, processing",
                "req": "object",
                "result": "object"
            }
        }
        //{"4ad8de25-f2ff-41d6-8783-86feef39dbc1":{"ts":1603906995024,"state":"cancelled","req":{"type":"purchase","amount":"1000","title":"Hello"},"result":{"ok":true,"error":"","terminal":{"returnCode":0,"transactionType":1,"transactionResultCode":7,"transactionResultCodeText":"interrupted","transactionNumber":-100,"transactionDetails":{"authorizationCode":"","serverMessage":"","merchantID":"454500000000000","maskedCardNumber":"","date":"2020.10.28","time":"18:42:06","cardsetName":"","readerUsed":"?","readerUsedText":"unknown","terminalIdentifier":"45450318","authorizationMessage":"000000","authorizationType":"?","authorizationTypeText":"unknown","transactionAmount":"","cardholderAuthorizationMethod":"?","cardholderAuthorizationMethodText":"no_authorization","atc":"","ac":"","aid":"","tvr":"","tsi":"","title":"","printingIndicator":"1","transactionCurrency":"PLN","terminalCurrency":"","exchangeRate":"","dcc1":"","dcc2":"","cashAmount":"","spendingAmount":"","dcc_currencyExp":"","additionalInformation":"DF0B0400000100"}}}},"bb9753cb-05fc-43dd-84a2-9af098bcc2b9":{"ts":1603907062507,"state":"processing","req":{"type":"purchase","amount":"1000","title":"Hello"},"result":{"ok":true,"error":"","terminal":{"returnCode":0,"transactionStatus":1,"transactionStatusText":"WaitCard","sequenceNb":"192"}}}}
    },
    "/v1/ingenico_transaction": {
        mapto: "/eservice/v1/ingenico_transaction",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_transaction": {
        method: 'post',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionStart(req, res); },
        schema: Joi.object().keys({
            type: Joi.string().required().valid('purchase', 'preauthorization', 'completion', 'preauthorization_cancel', 'refund', 'reversal', 'giftcard_activation', 'giftcard_balance', 'giftcard_topup', 'reconciliation', 'connection_test', 'tms'),
            amount: Joi.number().required(),
            title: Joi.string(),
            sequenceNb: Joi.string().required(),
            arguments: Joi.object().keys({
                transactionFlags: Joi.array().items(Joi.string().valid('TokenGeneration' /*0x01*/, 'ReferenceNumberGeneration'/*0x02*/, 'TransitTransactionIndicator'/*0x04*/)),     //DF01
                blikCode: Joi.number(),             //DF02 – BLIK code
                additionalPrintout: Joi.string(),   //DF03 – Additional printout data
                hints: Joi.string(),                //DF04 – "Hints"
                rfu: Joi.string(),                  //DF05 – RFU
                variableSymbol: Joi.string(),       //DF06 – Variable Symbol
                specificSymbol: Joi.string(),       //DF07 – Specific Symbol
                saleItemList: Joi.array().items(Joi.object().keys({   //7F01 – Sale Item List
                    itemId: Joi.string(),               //DF01 - Item ID 
                    productCode: Joi.string(),          //DF02 – Product Code 
                    amount: Joi.number(),               //DF03 - Amount(Total cost of item)
                    unitMeasure: Joi.string(),          //DF04 – Unit Measure
                    unitPrice: Joi.number(),            //DF05 – Unit Price(Price of single unit of item) 
                    quantity: Joi.number(),             //DF06 - Quantity 
                    taxCode: Joi.number()               //DF07 – Tax Code
                }) )
            })
        }),
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string",
        },
        description: "sequenceNb is mandatory but for completion and reversal types only. Set it to '-' for other types"
    },
    "/polcard/v1/ingenico_transaction": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoTransactionStart(req, res); },
        schema: apipolcard.ingenicoTransactionSchema(),
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string",
        },
        description: ""
    },
    "/v1/ingenico_transaction_last/:transactionIndex": {
        mapto: "/eservice/v1/ingenico_transaction_last/:transactionIndex",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_transaction_last/:transactionIndex": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionLast(req, res); }
    },
    "/v1/ingenico_transaction_end": {
        mapto: "/eservice/v1/ingenico_transaction_end",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_transaction_end": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionEnd(req, res); },
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string",
            "terminal": {
                "returnCode": "number: (0,1,2,7) - 0 accepted, 1 refused, 2 No connection, 7 interrupted by the user",
                "transactionType": "number",
                "transactionResultCode": "number",
                "transactionResultCodeText": "string",
                "transactionNumber": "number",
                "transactionDetails": {
                    "authorizationCode": "string",
                    "serverMessage": "string",
                    "merchantID": "string: example '454500000000000'",
                    "maskedCardNumber": "string: example '************0211'",
                    "date": "string: example format 2021.01.25",
                    "time": "string: example format 12:59:32",
                    "cardsetName": "string: example 'VISA'",
                    "readerUsed": "string: example 'C' - possible values: B Contactless, M Manual, C Magstripe, P ICC, ? Unknown",
                    "readerUsedText": "string: possible values 'magstripe', 'contactless', 'manual', 'icc', 'unknown'",
                    "terminalIdentifier": "string: example '45450318'",
                    "authorizationMessage": "string: example '000180'",
                    "authorizationType": "string: possible values 1 - online, 3 – offline, 4 – referral",
                    "authorizationTypeText": "string: possible values 'online', 'offline', 'referral'",
                    "transactionAmount": "string",
                    "cardholderAuthorizationMethod": "string: possible values A – PIN, @ - Signature, B – PIN and Signature, ? – no authorization",
                    "cardholderAuthorizationMethodText": "string: possible values 'pin', 'signature', 'pin_signature', 'no authorization', 'unknown'",
                    "atc": "string",
                    "ac": "string",
                    "aid": "string",
                    "tvr": "string",
                    "tsi": "string",
                    "title": "string",
                    "printingIndicator": "string: example '1'",
                    "transactionCurrency": "string: example 'PLN'",
                    "terminalCurrency": "string",
                    "exchangeRate": "string",
                    "dcc1": "string",
                    "dcc2": "string",
                    "cashAmount": "string",
                    "spendingAmount": "string",
                    "dcc_currencyExp": "string",
                    "additionalInformation": "string: example 'DF060A534E204E722056697361DF0B0400000100' or 'DF0B0400000100' or 'DF060A534E204E722056697361DF0B0400000400DF1318343534353033313830303031323630313231323132383534'",
                    "additionalInformationObject": "object"
                },
                "error": "string"
            }
        }
    },
    "/polcard/v1/ingenico_transaction_end": {
        method: 'get',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processPolcardIngenicoTransactionEnd(req, res); }
    },
    "/polcard/v1/ingenico_transaction_last": {
        method: 'get',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processPolcardIngenicoTransactionLast(req, res); }
    },
    "/polcard/v1/ingenico_transaction_abort": {
        method: 'get',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processPolcardIngenicoTransactionAbort(req, res); }
    },
    "/v1/ingenico_batch_recon_data/:transactionIndex": {
        mapto: "/eservice/v1/ingenico_batch_recon_data/:transactionIndex",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_batch_recon_data/:transactionIndex": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionReconData(req, res); }
    },
    "/v1/ingenico_batch_recon_info": {
        mapto: "/eservice/v1/ingenico_batch_recon_info",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_batch_recon_info": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionReconInfo(req, res); },
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string",
            "terminal": {
                "returnCode": "number: (0,1,2,7) - 0 accepted, 1 refused, 2 No connection, 7 interrupted by the user",
                "error": "string",
                "results": "[] array"
            }
        }
    },
    "/v1/ingenico_transaction_data_input": {
        mapto: "/eservice/v1/ingenico_transaction_data_input",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_transaction_data_input": {
        method: 'post',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionDataInput(req, res); },
        schema: Joi.object().keys({
            text: Joi.string().required(),
            sequenceNb: Joi.string().required()
        }),
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string"
        }
    },
    "/polcard/v1/ingenico_transaction_data_input": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoTransactionDataInput(req, res); },
        schema: Joi.object().keys({
            text: Joi.string().required(),
            type: Joi.number().required()
        })
    },
    "/v1/ingenico_transaction_send_key": {
        mapto: "/eservice/v1/ingenico_transaction_send_key",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_transaction_send_key": {
        method: 'post',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionSendKey(req, res); },
        schema: Joi.object().keys({
            keycode: Joi.string().required().valid('ok', 'back', 'cancel'),
            sequenceNb: Joi.string().required()
        }),
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string"
        }
    },
    "/polcard/v1/ingenico_transaction_send_key": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoTransactionSendKey(req, res); },
        schema: Joi.object().keys({
            keycode: Joi.string().required().valid('ok', 'back', 'cancel', 'yes', 'no', 'anykey')
        })
    },
    "/polcard/v1/ingenico_transaction_send_code": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoTransactionSendCode(req, res); },
        schema: Joi.object().keys({
            code: Joi.string().required(),
            type: Joi.string().required().valid('commercial', 'authorization')
        })
    },
    "/polcard/v1/ingenico_transaction_send_amount": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoTransactionSendAmount(req, res); },
        schema: Joi.object().keys({
            amount: Joi.number().required(),
            type: Joi.string().required().valid('cashback', 'other')
        })
    },
    "/polcard/v1/ingenico_transaction_send_option": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoTransactionSendOption(req, res); },
        schema: Joi.object().keys({
            option: Joi.number().required()
        })
    },
    "/polcard/v1/ingenico_transaction_send_profile": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoTransactionSendProfile(req, res); },
        schema: Joi.object().keys({
            profile: Joi.number().required()
        })
    },
    "/polcard/v1/ingenico_print": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoPrint(req, res); },
        schema: Joi.object().keys({
            messageNumber: Joi.number().required(),
            transport: apipolcard.ingenicoTransportSchema(),
            printout: Joi.array().required().items(Joi.object().keys({
                type: Joi.string().required().valid("header", "body", "footer", "date", "time", "logo"),
                align: Joi.string().required().valid("left", "right", "center"),
                bold: Joi.boolean().default(false),
                high: Joi.boolean().default(false),
                expand: Joi.boolean().default(false),
                text: Joi.string()
            }))
        })
    },
    "/polcard/v1/ingenico_prompt": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoPrompt(req, res); },
        schema: Joi.object().keys({
            transport: apipolcard.ingenicoTransportSchema().required(),
            type: Joi.string().required().valid("profiles", "satisfaction", "loyalty_auth", "last_upload_status", "additional_data_keypad" /*, "barcode"*/),
            reqestParams: Joi.string().allow('')
        })
    },
    "/polcard/v1/ingenico_prompt_barcode": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoPromptBarcode(req, res); },
        schema: Joi.object().keys({
            transport: apipolcard.ingenicoTransportSchema().required(),
            barcode: Joi.string().required(),
            headerText: Joi.string().required(),
            bodyText: Joi.string().required(),
            time: Joi.number().required()
        })
    },
    "/polcard/v1/ingenico_barcode": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoBarcode(req, res); },
        schema: Joi.object().keys({
            action: Joi.string().required().valid("scan_one", "scan_many", "finish")
        })
    },
    "/polcard/v1/ingenico_send_action": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoSetAction(req, res); },
        schema: Joi.object().keys({
            transport: apipolcard.ingenicoTransportSchema().required(),
            action: Joi.string().required().valid("download", "restart")
        })
    },
    "/polcard/v1/ingenico_cutoff": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoCutoff(req, res); },
        schema: Joi.object().keys({
            transport: apipolcard.ingenicoTransportSchema().required()
        })
    },
    "/polcard/v1/ingenico_cutoff_report": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoCutoffReport(req, res); },
        schema: Joi.object().keys({
            transport: apipolcard.ingenicoTransportSchema().required()
        })
    },
    "/polcard/v1/ingenico_cutoff_end": {
        method: 'get',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoCutoffEnd(req, res); }
    },
    "/polcard/v1/ingenico_report": {
        method: 'post',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoReport(req, res); },
        schema: Joi.object().keys({
            transport: apipolcard.ingenicoTransportSchema().required(),
            reportType: Joi.string().required().valid("batch_report", "cumulative_report"),
            action: Joi.string().required().valid("batch_return", "manual_batch_then_return", "cumulative_return", "cumulative_return_reset_counters"),
            profileId: Joi.number().required().min(0).max(99),
            timeout: Joi.number().required()
        })
    },
    "/polcard/v1/ingenico_report_end": {
        method: 'get',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoReportEnd(req, res); }
    },
    // "/polcard/v1/ingenico_listen": {
    //     method: 'get',
    //     grouptag: 'polcard',
    //     func: (req, res) => { apipolcard.processIngenicoListen(req, res); }
    // },
    "/polcard/v1/ingenico_prompt_end": {
        method: 'get',
        grouptag: 'polcard',
        func: (req, res) => { apipolcard.processIngenicoPromptEnd(req, res); }
    },
    "/v1/ingenico_set_parameter": {
        mapto: "/eservice/v1/ingenico_set_parameter",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_set_parameter": {
        method: 'post',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTerminalSetParameter(req, res); },
        schema: Joi.object().keys({
            type: Joi.string().required().valid('volatile', 'permanent'),
            parameters: Joi.array().required().items(Joi.object().keys({
                key: Joi.string().required(),
                value: Joi.string().required()
            }))
        })
    },
    "/v1/ingenico_set_standby": {
        mapto: "/eservice/v1/ingenico_set_standby",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_set_standby": {
        method: 'post',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTerminalSetStandby(req, res); },
        schema: Joi.object().keys({
            timeoutSeconds: Joi.number().required()
        })
    },
    "/v1/ingenico_report_generate": {
        mapto: "/eservice/v1/ingenico_report_generate",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_report_generate": {
        method: 'post',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionReportGenerate(req, res); }
    },
    "/v1/ingenico_report_details/:cardset": {
        mapto: "/eservice/v1/ingenico_report_details/:cardset",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_report_details/:cardset": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionGetSummary(req, res); }
    },
    "/v1/ingenico_report_summary": {
        mapto: "/eservice/v1/ingenico_report_summary",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_report_summary": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionGetTotalSummary(req, res); }
    },

    //------------------------------------------------------------------------------------------------------------------
    "/v1/ingenico_analysis_generate": {
        mapto: "/eservice/v1/ingenico_analysis_generate",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_analysis_generate": {
        method: 'post',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionAnalysisGenerate(req, res); }
    },
    "/v1/ingenico_analysis_details/:transactionIndex": {
        mapto: "/eservice/v1/ingenico_analysis_details/:transactionIndex",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_analysis_details/:transactionIndex": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionAnalysisDetails(req, res); }
    },
    "/v1/ingenico_analysis_summary": {
        mapto: "/eservice/v1/ingenico_analysis_summary",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_analysis_summary": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionAnalysisTotalSummary(req, res); }
    },

    //------------------------------------------------------------------------------------------------------------------
    "/v1/ingenico_dcc_analysis_generate": {
        mapto: "/eservice/v1/ingenico_dcc_analysis_generate",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_dcc_analysis_generate": {
        method: 'post',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionDCCAnalysisGenerate(req, res); }
    },
    "/v1/ingenico_dcc_analysis_details/:transactionIndex": {
        mapto: "/eservice/v1/ingenico_dcc_analysis_details/:transactionIndex",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_dcc_analysis_details/:transactionIndex": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionDCCAnalysisDetails(req, res); }
    },
    "/v1/ingenico_dcc_analysis_summary": {
        mapto: "/eservice/v1/ingenico_dcc_analysis_summary",
        grouptag: 'eservice',
        deprecated: true
    },
    "/eservice/v1/ingenico_dcc_analysis_summary": {
        method: 'get',
        grouptag: 'eservice',
        func: (req, res) => { processIngenicoTransactionDCCAnalysisTotalSummary(req, res); }
    },

    //------------------------------------------------------------------------------------------------------------------
    "/": {
        method: 'get',
        func: (req, res) => { processRootRestApi(req, res); }
    }
}

function convertToSampleJSON(schema) {
    var result = {};
    if (schema['$_terms'].keys != null) {
        schema['$_terms'].keys.forEach(key => {
            //console.log('aaaaaaaaaaaaaaaaaa:' + key.key);

            if (key.schema.type == 'object') {
                //console.log('--------------------');
                //console.log(key.schema);
                var rrr = JSON.stringify(convertToSampleJSON(key.schema));
                //console.log('got --------------------' + rrr);
                result[key.key] = rrr;
            } else {
                result[key.key] = key.schema.type;
                if (key.schema['_valids'] != null && key.schema['_valids']['_values'] != null) {
                    //console.log(key.schema['_valids']['_values']);
                    result[key.key] += ': ';
                    const array = [...key.schema['_valids']['_values']];
                    result[key.key] += array.join(', ');
                }
            }
        });
    }
    return result;
}

function getProtocol(req) {
    var proto = req.connection.encrypted ? 'https' : 'http';
    proto = req.headers['x-forwarded-proto'] || proto;
    return proto.split(/\s*,\s*/)[0];
}

function initOtherOptions(other_options) {
    buildVersion = other_options.buildVersion;
    schemaVersion = other_options.schemaVersion;
}


function getCashRegisterNumber(req) {
    if (req == null) return 1;
    if (req.query == null) return 1;
    if (req.query.cashRegisterNumber == null) return 1;
    if (isNaN(req.query.cashRegisterNumber)) return 1;
    return req.query.cashRegisterNumber;
}

function isResetFlagEnabled(req) {
    if (req == null) return false;
    if (req.query == null) return false;

    if (req.query.reset == null) return false;
    if (req.query.reset == "false") return false;

    if (req.query.reset == "true") return true;

    return false;

    //return req.query.reset != null ? (req.query.reset ? true : false) : false;
}
function initApiDefinitions(def) {
    for (const [key, value] of Object.entries(def)) {
        const targetvalue = (value.mapto != null ? def[value.mapto] : value);
        if (targetvalue.method == 'get')
            app.get(key, targetvalue.func);
        else if (targetvalue.method == 'put')
            app.put(key, targetvalue.func);
        else if (targetvalue.method == 'post')
            app.post(key, targetvalue.func);
        else if (targetvalue.method == 'delete')
            app.delete(key, targetvalue.func);
        else
            logger.logger.error('Unrecognized method: ' + targetvalue.method);
    }
}

function initHiddenApiDefinitions() {
    app.get('/v1/tests/run', (req, res) => { processRunTests(req, res) });
    app.post('/v1/tests/log', (req, res) => { processLogExtra(req, res) });
}

function processRootRestApi(req, res) {
    var restpoints = [
        { groupname: 'General information' },
        { type: 'GET', name: '' },
        { groupname: 'License' }
    ];
    for (const [key, value] of Object.entries(api_definition)) {
        if (value.grouptag == 'license')
            restpoints.push({ type: value.method.toUpperCase(), name: key.substr(1), deprecated: false });
    }

    restpoints.push({ groupname: 'IngenicoServer methods (eService)' });
    for (const [key, value] of Object.entries(api_definition)) {
        if (key != "/" && key.startsWith("/auto_queue") == false && value.grouptag == 'eservice') {
            if (value.mapto == null)
                restpoints.push({ type: value.method.toUpperCase(), name: key.substr(1), deprecated: false });
            else
                restpoints.push({ type: api_definition[value.mapto].method.toUpperCase(), name: key.substr(1), deprecated: value.deprecated });
        }
    }

    restpoints.push({ groupname: 'IngenicoServer methods (Polcard - First Data)' });
    for (const [key, value] of Object.entries(api_definition)) {
        if (key != "/" && key.startsWith("/auto_queue") == false && value.grouptag == 'polcard') {
            if (value.mapto == null)
                restpoints.push({ type: value.method.toUpperCase(), name: key.substr(1), deprecated: false });
            else
                restpoints.push({ type: api_definition[value.mapto].method.toUpperCase(), name: key.substr(1), deprecated: value.deprecated });
        }
    }
    /*if (req.query.textdetails) {
        restpoints.push({ groupname: 'IngenicoServer methods (details)' });
        for (const [key, value] of Object.entries(api_definition)) {
            if (key != "/" && key.startsWith("/auto_queue") == false)
                restpoints.push({ type: value.method.toUpperCase(), name: key.substr(1), schema: value.schema, schema_resp: value.schema_resp, description: value.description });
        }
    }*/

    var restpointsstring = '';    
    restpoints.forEach(rp => {
        if (rp.deprecated) {
            // do nothing
        } else if (rp.groupname != null) {
            restpointsstring += '<br>\n<b>' + rp.groupname + '</b><br>\n';
        } else if (rp.ref != null) {
            restpointsstring += api_definition[rp.ref].method.toUpperCase().padEnd(6, ' ').replace(/ /g, '&nbsp;') + ' ' + getProtocol(req) + '://' + req.headers.host + '/' + rp.ref.substr(1) + '  <br>\n';
            if (api_definition[rp.ref].attributes != null && api_definition[rp.ref].attributes.length > 0) {
                api_definition[rp.ref].attributes.forEach(att => {
                    restpointsstring += '       '.replace(/ /g, '&nbsp;') + ' ?' + att.attribute + ' = ' + att.description + ' (default: ' + att.default + ')  <br>\n';
                });
            }
        } else {
            restpointsstring += rp.type.padEnd(6, ' ').replace(/ /g, '&nbsp;') + ' ' + getProtocol(req) + '://' + req.headers.host + '/' + rp.name + '  <br>\n';

            if (rp.schema)
                restpointsstring += 'Request:<br>\n' + apitools.prettyJson(convertToSampleJSON(rp.schema)).replace(/\t/g, '<br>') + '  <br>\n';

            if (rp.schema_resp)
                restpointsstring += 'Response:<br>\n' + apitools.prettyJson(rp.schema_resp).replace(/\t/g, '<br>') + '  <br>\n';

            if (rp.description)
                restpointsstring += 'Description:<br>\n' + (rp.description) + '  <br>\n';

            if (rp.schema || rp.schema_resp)
                restpointsstring += '  <hr>\n';
            //restpointsstring += '  <hr>\n';
        }
    });

    restpointsstring += '<br>\n<b>Deprecated methods (compatible with 1.3)</b><br>\n';
    restpoints.forEach(rp => {
        if (rp.deprecated) {
            restpointsstring += rp.type.padEnd(6, ' ').replace(/ /g, '&nbsp;') + ' ' + getProtocol(req) + '://' + req.headers.host + '/' + rp.name + '  <br>\n';
        }
    });
    //if (!req.query.textdetails) {
        //restpointsstring += '<br>\n<a href="/?textdetails=true">Show</a><br>\n';  //be zsensu dluzej to pokazywac skro calal dokumentacja i tak jest w swagger
    //}

    res.send(
        '<tt>\n' +
        'Ingenico Server v.' + buildVersion + ' (schema ' + schemaVersion + ') © 2021 BigDotSoftware (bigdotsoftware@bigdotsoftware.pl). <br>\n' +
        'How to start: https://blog.bigdotsoftware.pl/ingenico-server-pierwsze-uzycie/ <br>\n<br>\n' +
        'Swagger YAML: &nbsp; <a href="https://github.com/bigdotsoftware/ingenicoserver/blob/main/swagger/swagger-' + schemaVersion + '.yaml">https://github.com/bigdotsoftware/ingenicoserver/blob/main/swagger/swagger-' + buildVersion + '.yaml</a> <br>\n<br>\n' +
        'Swagger Editor: <a href="https://editor.swagger.io/?url=https://download2.bigdotsoftware.pl/github_ingenico_swagger/swagger-' + schemaVersion + '.yaml">Open in Editor</a> <br>\n<br>\n' +
        'RESTful API: <br></tt><tt>\n' +
        restpointsstring + 
        '</tt>'
    );
}

function replaceAll(string, search, replace) {
    return string.split(search).join(replace);
}


function processLicenseUpgrade(req, res) {
    setTimeout(function () {

        if (config.license.file != null && config.license.file != '') {
            logger.logger.info('license file: ' + config.license.file);
            var stat = nativeTest.ValidateLicenseFile({ file: config.license.file });
            logger.logger.info(JSON.stringify(stat));
            if (stat.results != null) {
                var uniqueLicenseId = '';
                stat.results.forEach(result => {
                    if (result.uniqueLicenseId != null && result.uniqueLicenseId != '') {
                        uniqueLicenseId = result.uniqueLicenseId;
                    }
                });
                if (uniqueLicenseId != '')
                    apitools.obtainNewLicenseFile(uniqueLicenseId);
                else
                    logger.logger.error('Cannot determine unique license ID');

                if (stat.ok) {
                    var statinfo = apitools.readStatInfo(stat);
                    return apitools.sendResponse(req, res, {
                        ok: true,
                        text: 'Wersja licencjonowana dla ' + statinfo.licenseTo + ' wazna do ' + new Date(statinfo.validTo * 1000).toISOString(),
                        validTo: statinfo.validTo,
                        licenseTo: statinfo.licenseTo,
                        expired: statinfo.expired,
                        requested: true,
                        requestedtext: 'Requested for new license file. Check logs for more details'
                    });
                } else {
                    return apitools.sendErrorResponse(req, res, 'Wersja nielicencjonowana');
                }
            }
        } else {
            return apitools.sendErrorResponse(req, res, 'Wersja nielicencjonowana');
        }

    }, 500);
}

function processLicenseCheck(req, res) {
    setTimeout(function () {
        var stat = apitools.CheckLicense();
        if (stat.ok) {
            var statinfo = apitools.readStatInfo(stat);

            if (req.query.json)
                return apitools.sendResponse(req, res, {
                    ok: true,
                    text: 'Wersja licencjonowana dla ' + statinfo.licenseTo + ' wazna do ' + new Date(statinfo.validTo * 1000).toISOString(),
                    validTo: statinfo.validTo,
                    licenseTo: statinfo.licenseTo,
                    expired: statinfo.expired
                });
            else
                return apitools.sendResponse(req, res, 'Wersja licencjonowana dla ' + statinfo.licenseTo + ' wazna do ' + new Date(statinfo.validTo * 1000).toISOString());
        } else {
            if (req.query.json)
                return apitools.sendErrorResponse(req, res, {
                    ok: false,
                    text: 'Wersja nielicencjonowana',
                    expired: true
                });
            else
                return apitools.sendErrorResponse(req, res, 'Wersja nielicencjonowana');
        }
    }, 500);
}

function processRunTests(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    const msg0 = nativeTest.RunTests(fulldebug ? "1" : "0");
    return apitools.sendResponse(req, res, msg0);
}

function processLogExtra(req, res) {

    const schema = Joi.object().keys({
        text: Joi.string().required(),
        level: Joi.string().required().valid('info','error','debug','warn')
    });
    var test = schema.validate(req.body);
    if (test.error != null) {
        logger.logger.error('Payload validation error:' + JSON.stringify(test));
        return apitools.sendErrorResponse(req, res, test.error);
    } else {
        switch (req.body.level) {
            case 'info':
                logger.logger.info(req.body.text);
                break;
            case 'error':
                logger.logger.error(req.body.text);
                break;
            case 'debug':
                logger.logger.debug(req.body.text);
                break;
            case 'warn':
                logger.logger.warn(req.body.text);
                break;
        }
        return res.json({});
    }
}
function processIngenicoStatus(req, res) {

    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        const msg0 = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber } );

        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);
    });


    /*

    var fulldebug = true;
    var stat = nativeTest.GeneralInformation(fulldebug ? "1" : "0");
    if (!stat.ok) {
        logger.logger.error(JSON.stringify(stat));
        return apitools.sendErrorResponse(req, res, stat.error);
    }
    //finalresults.ok = true;
    return apitools.sendResponse(req, res, {});
    */
    
}

function processIngenicoGeneralInfo(req, res) {

    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        const msg0 = nativeTest.GetTermGeneralInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber } );

        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);
    });
}

function verifyTermResponse(req, res, result, uuid /*optional*/) {
    if (result.terminal.returnCode != 0) {
        if (res != null )
            apitools.sendErrorResponse(req, res, result);

        var closeresult = apitools.ConnectionEnd();
        /*if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);*/
        if (uuid != null)
            setRequestError(uuid, result.error, result);

        return false;
    } else {
        return true;
    }
}

function processIngenicoTransactionReconData(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    const transactionIndex = req.params.transactionIndex;
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const trReconData = nativeTest.TransactionReconData(fulldebug ? "1" : "0", { transactionIndex: transactionIndex, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trReconData)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, trReconData);
    });
}

function processIngenicoTransactionLast(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    const transactionIndex = req.params.transactionIndex;
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const trReconData = nativeTest.TransactionLastData(fulldebug ? "1" : "0", { transactionIndex: transactionIndex, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trReconData)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, trReconData);
    });
}


function processIngenicoTransactionReconInfo(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const trReconInfo = nativeTest.TransactionReconInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trReconInfo)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, trReconInfo);
    });
}

function processIngenicoTransactionDataInput(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    if (!apitools.validateSchema(req, res, api_definition))
        return;

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const trReconInfo = nativeTest.TransactionDataInput(fulldebug ? "1" : "0", { text: req.body.text, sequenceNb: req.body.sequenceNb, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trReconInfo)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, trReconInfo);
    });
}

function processIngenicoTransactionSendKey(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    if (!apitools.validateSchema(req, res, api_definition))
        return;

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        var keycode = '';
        if (req.body.keycode == 'ok')
            keycode = "\\n"
        else if (req.body.keycode == 'back')
            keycode = "\\b"
        else if (req.body.keycode == 'cancel')
            keycode = "\\e"

        const trReconInfo = nativeTest.TransactionDataInput(fulldebug ? "1" : "0", { text: keycode, sequenceNb: req.body.sequenceNb, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trReconInfo)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, trReconInfo);
    });
}


function processIngenicoTerminalSetParameter(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    if (!apitools.validateSchema(req, res, api_definition))
        return;

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TerminalSetParameter(fulldebug ? "1" : "0", {
            parameters: req.body.parameters,
            typeOfParameter: (req.body.type == 'permanent' ? 1 : 0),    //('volatile', 'permanent')
            cashRegisterNumber: cashRegisterNumber
        });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);
    });
}


function processIngenicoTerminalSetStandby(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    if (!apitools.validateSchema(req, res, api_definition))
        return;

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TerminalSetStandby(fulldebug ? "1" : "0", {
            timeoutSeconds: req.body.timeoutSeconds,
            cashRegisterNumber: cashRegisterNumber
        });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);
    });
}



//-----------------------------------------------------------------------------
function processIngenicoTransactionReportGenerate(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    var reset = isResetFlagEnabled(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionReportGenerate(fulldebug ? "1" : "0", { reset: (reset ? "1" : "0"), cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);
    });
}

function processIngenicoTransactionGetSummary(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    const cardset = req.params.cardset;
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionReportGetSummary(fulldebug ? "1" : "0", { cardset: cardset, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);
    });
}

function processIngenicoTransactionGetTotalSummary(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionReportGetTotalSummary(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);
    });
}

//-----------------------------------------------------------------------------
function processIngenicoTransactionAnalysisGenerate(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionAnalysisGenerate(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);
    });
}

function processIngenicoTransactionAnalysisDetails(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    const transactionIndex = req.params.transactionIndex;
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionAnalysisGetDetails(fulldebug ? "1" : "0", { transactionIndex: transactionIndex, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);
    });
}

function processIngenicoTransactionAnalysisTotalSummary(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionAnalysisGetTotalSummary(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);
    });
}

//-----------------------------------------------------------------------------
function processIngenicoTransactionDCCAnalysisGenerate(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionDCCAnalysisGenerate(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);
    });
}

function processIngenicoTransactionDCCAnalysisDetails(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    const transactionIndex = req.params.transactionIndex;
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionDCCAnalysisGetDetails(fulldebug ? "1" : "0", { transactionIndex: transactionIndex, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);
    });
}

function processIngenicoTransactionDCCAnalysisTotalSummary(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionDCCAnalysisGetTotalSummary(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, msg0);
    });
}

function ClearBatchCompleted(req, res, fulldebug) {
    var cashRegisterNumber = getCashRegisterNumber(req);

    for (var i = 0; i < 100; i++) {
        const trReconInfo = nativeTest.TransactionReconInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trReconInfo)) return false;

        var trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trStatus)) return false;

        if (trStatus.terminal.transactionStatus == 0)
            break;
    }
    return true;
}

function processIngenicoTransactionEnd(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        var trEnd = nativeTest.TransactionEnd(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trEnd)) return;
        logger.logger.info('trEnd done: ' + JSON.stringify(trEnd));
        
        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, trEnd);
    });
}

function sleep(ms) {
    return new Promise((resolve) => {
        setTimeout(resolve, ms);
    });
}

function processIngenicoTransactionCancel(req, res) {
    //var fulldebug = apitools.isFullDebugMode(req);
    var uuid = req.params.uuid;

    if (apiasyncreqs.getCurrentRequests()[uuid] != null) {
        logger.logger.info(`Marking request ${uuid} cancelled`);
        setRequestCancelled(uuid, { ok: false });

        return apitools.sendResponse(req, res, apiasyncreqs.getCurrentRequests()[uuid]);
    } else {
        return apitools.sendErrorResponse(req, res, 'Wrong UUID');
    }
}

function processIngenicoTransactionStatus(req, res) {
    if (apiasyncreqs.getCurrentRequests()[req.params.uuid] != null)
        return apitools.sendResponse(req, res, apiasyncreqs.getCurrentRequests()[req.params.uuid]);
    else
        return apitools.sendErrorResponse(req, res, 'Wrong UUID');

}
function processIngenicoTransactionStatuses(req, res) {
    return apitools.sendResponse(req, res, apiasyncreqs.getCurrentRequests());
}
function processIngenicoTransactionAuto(req, res) {
    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    if (!apitools.validateSchema(req, res, api_definition))
        return;

    apitools.ConnectionBegin(req, async (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        const transactionUUID = uuid();
        setRequestProcessing(transactionUUID, req.body);

        //------------------------
        var trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trStatus, transactionUUID)) return;
        logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));

        if (trStatus.terminal.transactionStatus == 129 /* 0x81 TRANSACTION_STATUS_RECONCILIATION_NEEDED*/) {
            const trReconciliation = nativeTest.TransactionBegin(fulldebug ? "1" : "0", {
                type: 'reconciliation',
                amount: 0,
                title: '',
                sequenceNb: '',
                arguments: {},
                cashRegisterNumber: cashRegisterNumber,
                file: apitools.GetLicenseFile()
            });
            logger.logger.info('trReconciliation done: ' + JSON.stringify(trReconciliation));
            if (!verifyTermResponse(req, res, trReconciliation, transactionUUID)) return;

            trStatus = await startListeningForWaitTrEnd(req, res, fulldebug, transactionUUID);

            var trEnd = nativeTest.TransactionEnd(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
            if (!verifyTermResponse(req, res, trEnd, transactionUUID)) return;
            logger.logger.info('trEnd done: ' + JSON.stringify(trEnd));

            trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
            if (!verifyTermResponse(req, res, trStatus, transactionUUID)) return;
            logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));
        }

        if (trStatus.terminal.transactionStatus == 13 /* 0x0D TRANSACTION_STATUS_BATCH_COMPLETED*/) {
            var rc = ClearBatchCompleted(req, res, fulldebug);
            if (!rc) {
                apitools.ConnectionEnd();
                return;
            }
        }


        if (trStatus.terminal.transactionStatus == 1 /* 0x01 TRANSACTION_STATUS_WAIT_CARD*/) {
            logger.logger.warn("Cannot interrupt transaction in WAIT_CARD state");

            apitools.ConnectionEnd();
            apitools.sendErrorResponse(req, res, { ok: false, error: trStatus.terminal.transactionStatusText + ' : Przyłuż kartę lub anuluj na terminalu rozpoczętą transakcję' });
            return;
        }

        if (trStatus.terminal.transactionStatus == 6 /* 0x06 TRANSACTION_STATUS_WAIT_TREND*/) {

            logger.logger.info("Interrupting current transaction in WAIT_TREND state");

            var trEnd = nativeTest.TransactionEnd(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
            if (!verifyTermResponse(req, res, trEnd, transactionUUID)) return;
            logger.logger.info('trEnd done: ' + JSON.stringify(trEnd));

            trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
            if (!verifyTermResponse(req, res, trStatus, transactionUUID)) return;
            logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));
        }

        //------------------------
        //var trEnd = nativeTest.TransactionEnd(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        //if (!verifyTermResponse(req, res, trEnd)) return;
        //logger.logger.info('trEnd done: ' + JSON.stringify(trEnd));

        //------------------------
        //trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        //if (!verifyTermResponse(req, res, trStatus)) return;
        //logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));

        //------------------------
        const trBegin = nativeTest.TransactionBegin(fulldebug ? "1" : "0", {
            type: req.body.type/* 'purchase'*/,   //purchase=1, preauthorization=4, completion=5, preauthorization_cancel=82, refund=6, reversal=10, giftcard_activation=42, giftcard_balance=43, giftcard_topup=44, reconciliation=31, connection_test=33, tms=40
            amount: req.body.amount != null ? req.body.amount : 0 /*1000*/,
            title: req.body.title != null ? req.body.title : '', //'Purchase 1234'
            sequenceNb: req.body.sequenceNb != null ? req.body.sequenceNb : '',
            arguments: req.body.arguments != null ? req.body.arguments : {},
            cashRegisterNumber: cashRegisterNumber,
            file: apitools.GetLicenseFile()
        });
        logger.logger.info('trBegin done: ' + JSON.stringify(trBegin));
        if (!verifyTermResponse(req, res, trBegin, transactionUUID)) return;

        //------------------------
        if (!trBegin.ok)
            apitools.sendErrorResponse(req, res, trBegin);
        else
            apitools.sendResponse(req, res, { ok: true, uuid: transactionUUID });

        trStatus = await startListeningForWaitTrEnd(req, null/*res - pas null to avoid sending response*/, fulldebug, transactionUUID);

        if (trStatus.terminal.transactionStatus != 6 /*WaitTrEnd*/) {
            logger.logger.info('Interrupting due to wrong status: ' + JSON.stringify(trStatus));
            apitools.ConnectionEnd();
            return;
        }

        var trEnd = nativeTest.TransactionEnd(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, null /*res - pas null to avoid sending response*/, trEnd, transactionUUID)) return;
        logger.logger.info('trEnd done: ' + JSON.stringify(trEnd));

        if (trEnd.terminal.transactionResultCode ==0 )
            setRequestSuccess(transactionUUID, trEnd);
        else if (trEnd.terminal.transactionResultCode == 1 || trEnd.terminal.transactionResultCode == 7)
            setRequestCancelled(transactionUUID, trEnd);
        else
            setRequestError(transactionUUID, "", trEnd);

        if (req.body.callback) {
            apitools.notifyBackendAboutTransactionResult(trEnd, req.body.callback);
        }

        if (config.posnetserver.active) {
            if (req.body.posnetserver) {
                if (trEnd.terminal.transactionResultCode == 0) {
                    apitools.sendToPosnetServer(req.body.posnetserver);
                } else {
                    logger.logger.warn(`Integracja z PosnetServer jest włączona, transakcja na terminalu nie zakonczyla sie sukcesem - paragon nie bedzie drukowany`);
                }
            } else {
                logger.logger.warn(`Integracja z PosnetServer jest włączona, ale request nie zawiera struktury 'posnetserver' którą można przekazać do serwisu.`);
            }
        }


        var result = apitools.ConnectionEnd();
        if (!result.ok) {
            logger.logger.error('cannot end: ' + JSON.stringify(result));
        }
    });
}


async function startListeningForWaitTrEnd(req, res, fulldebug, uuid) {

    var cashRegisterNumber = getCashRegisterNumber(req);

    var trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
    updateCurrentResponseObj(uuid, trStatus);
    if (!verifyTermResponse(req, res, trStatus)) return trStatus;
    logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));
    

    while (trStatus.terminal.transactionStatus != 6 /*WaitTrEnd*/) {
        trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        updateCurrentResponseObj(uuid, trStatus);
        if (!verifyTermResponse(req, res, trStatus)) return trStatus;
        logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));


        if (apiasyncreqs.getCurrentRequests()[uuid] != null && apiasyncreqs.getCurrentRequests()[uuid].state == 'cancelled') {
            logger.logger.info(`Received cancelled flag for ${uuid}`);
            return trStatus;
        }

        await sleep(1000);
    }

    return trStatus;
}

function processIngenicoTransactionStart(req, res) {

    var fulldebug = apitools.isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    if (!apitools.validateSchema(req, res, api_definition))
        return;

    apitools.ConnectionBegin(req, (err) => {
        if (!err.ok) {
            apitools.sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        var trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trStatus)) return;
        logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));

        if (trStatus.terminal.transactionStatus == 13 /* 0x0D TRANSACTION_STATUS_BATCH_COMPLETED*/) {
            var rc = ClearBatchCompleted(req, res, fulldebug);
            if (!rc) {
                apitools.ConnectionEnd();
                return;
            }
        }
        

        if (trStatus.terminal.transactionStatus == 1 /* 0x01 TRANSACTION_STATUS_WAIT_CARD*/) {
            apitools.ConnectionEnd();
            apitools.sendErrorResponse(req, res, { ok: false, error: trStatus.terminal.transactionStatusText + ' : Przyłuż kartę lub zamknij rozpoczętą transakcję' });
            return;
        }
        if (trStatus.terminal.transactionStatus == 6 /* 0x06 TRANSACTION_STATUS_WAIT_TREND*/) {
            apitools.ConnectionEnd();
            apitools.sendErrorResponse(req, res, { ok: false, error: trStatus.terminal.transactionStatusText + ' : Zamknij niedokończoną transakcję'});
            return;
        }

        //------------------------
        //var trEnd = nativeTest.TransactionEnd(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        //if (!verifyTermResponse(req, res, trEnd)) return;
        //logger.logger.info('trEnd done: ' + JSON.stringify(trEnd));

        //------------------------
        //trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        //if (!verifyTermResponse(req, res, trStatus)) return;
        //logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));

        //------------------------
        logger.logger.info('TransactionBegin - preparing');
        const trBegin = nativeTest.TransactionBegin(fulldebug ? "1" : "0", {
            type: req.body.type/* 'purchase'*/,   //purchase=1, preauthorization=4, completion=5, preauthorization_cancel=82, refund=6, reversal=10, giftcard_activation=42, giftcard_balance=43, giftcard_topup=44, reconciliation=31, connection_test=33, tms=40
            amount: req.body.amount != null ? req.body.amount : 0 /*1000*/,
            title: req.body.title != null ? req.body.title : '', //'Purchase 1234'
            sequenceNb: req.body.sequenceNb != null ? req.body.sequenceNb : '',
            arguments: req.body.arguments != null ? req.body.arguments : {},
            cashRegisterNumber: cashRegisterNumber,
            file: apitools.GetLicenseFile()
        });
        logger.logger.info('trBegin done: ' + JSON.stringify(trBegin));
        if (!verifyTermResponse(req, res, trBegin)) return;


        //------------------------
        var result = apitools.ConnectionEnd();
        if (!result.ok)
            return apitools.sendErrorResponse(req, res, result);
        else
            return apitools.sendResponse(req, res, {ok: true});
    });
}


function initNative() {

    //var useddriver = config.posnet.driver != null ? config.posnet.driver : 'posnet';
    var usedcommanddelay = config.ingenico.commanddelay != null ? config.ingenico.commanddelay : 200;
    var usedwaitdelay = config.ingenico.waitdelay != null ? config.ingenico.waitdelay : 100;
    var usedcommandresponsetimeout = config.ingenico.commandresponsetimeout != null ? config.ingenico.commandresponsetimeout : 10000;
    var responsewaitdelay = config.ingenico.responsewaitdelay != null ? config.ingenico.responsewaitdelay : 100;

    logger.logger.info('Initializing ingenico module');
    nativeTest.InitializeModule({
        // loggerterminal: config.logger.terminal,
        loggernative: config.logger.native,
        licensefile: apitools.GetLicenseFile(),
        //posnetdriver: useddriver,
        commanddelay: usedcommanddelay,
        waitdelay: usedwaitdelay,
        responsewaitdelay: responsewaitdelay,
        commandresponsetimeout: usedcommandresponsetimeout
    });
}

function initScheduleQueueCleaner() {

    logger.logger.info(`Scheduling queue cleaner with retention of ${config.queue.retention} ms`);
    setInterval(function () {

        //const currentts = new Date().getTime();
        const retention = config.queue.retention;

        apiasyncreqs.cleanOlderThan(retention);

    }, 2000);
}

function initPinger() {
    if (config.ingenico.ping.active) {
        var cashRegisterNumber = config.ingenico.ping.cashRegisterNumber;
        if (cashRegisterNumber == null)
            cashRegisterNumber = 1;
        logger.logger.info(`Scheduling pinger with interval ${config.ingenico.ping.interval} ms on cashRegisterNumber=${cashRegisterNumber}`);
        setInterval(function () {

            var fulldebug = false;

            apitools.ConnectionBegin(null, (err) => {
                if (!err.ok) {
                    logger.logger.error('PING|Port open error: ' + JSON.stringify(err));
                    return;
                }

                const msg0 = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
                var txt = msg0.terminal.transactionStatusText != null ? msg0.terminal.transactionStatusText : '';
                logger.logger.info(`PING|status: ${msg0.terminal.transactionStatus} with text: '${txt}'`);

                var result = apitools.ConnectionEnd();
                if (!result.ok)
                    logger.logger.error('PING|Port close error: '+ JSON.stringify(result) );
            });

        }, config.ingenico.ping.interval);
    } else {
        logger.logger.info(`Pinger is disabled in configuration. To enable it set ingenico.ping.active=true`);
    }
}



function initLicense() {
    logger.logger.info(`Checking license file ${apitools.GetLicenseFile()} .....`);
    var stat = apitools.CheckLicense();
    return stat;
}
function init(http_options, https_options, other_options) {

    initApiDefinitions(api_definition);
    initHiddenApiDefinitions();
    initOtherOptions(other_options);
    initScheduleQueueCleaner();

    console.log(' _                         _                ______                                ' );
    console.log('| |                       (_)              / _____)                               ' );
    console.log('| |____   ____ _____ ____  _  ____ ___    ( (____  _____  ____ _   _ _____  ____  ' );
    console.log('| |  _ \\ / _  | ___ |  _ \\| |/ ___) _ \\    \\____ \\| ___ |/ ___) | | | ___ |/ ___) ');
    console.log('| | | | ( (_| | ____| | | | ( (__| |_| |   _____) ) ____| |    \\ V /| ____| |     ' );
    console.log('|_|_| |_|\\___ |_____)_| |_|_|\\____)___/   (______/|_____)_|     \\_/ |_____)_|     ' );
    console.log('        (_____|                                                                   ' );
    logger.logger.info(`Trying to open HTTP on port: ${http_options.port}`);
    try {
        var server = http.createServer(app);
        server.on('error', function (e) {
            logger.logger.error(`Cannot (1) start HTTP at http://localhost:${http_options.port}, please check if there no another process listening on the same port. Reason: ${e}`);
        });
        server.listen(http_options.port/*, '127.0.0.1'*/);
        logger.logger.info('HTTP  at http://localhost:' + http_options.port)
    } catch (e) {
        logger.logger.error(`Cannot start HTTP at http://localhost:'${http_options.port}, please check if there no another process listening on the same port. Reason: ${e}`);
    }


    if (https_options.active) {
        logger.logger.info(`Loading SSL certs...: ${https_options.sslcertificates.key}`);
        var options = null;
        try {
            options = {
                key: fs.readFileSync(https_options.sslcertificates.key),
                cert: fs.readFileSync(https_options.sslcertificates.crt)
            };
        } catch (e) {
            logger.logger.error(`Cannot load certificates ${https_options.sslcertificates.key} and ${https_options.sslcertificates.crt}. Reason: ${e}`);
        }

        if (options != null) {
            logger.logger.info(`Trying to open HTTPS on port: ${https_options.port}`);
            try {
                https.createServer(options, app).listen(https_options.port);
                logger.logger.info(`HTTPS at https://localhost: ${https_options.port}`);
            } catch (e) {
                logger.logger.error(`Cannot start HTTPS at https://localhost: ${https_options.port}, please check your certificate under the ./cert directory. Reason: ${e}`);
            }
        }
    } else {
        logger.logger.info('HTTPS is disabled in configuration. To enable set https.active=true in config.yml');
    }

    initNative();
    initDrivers();
    initPinger();
    initLicense();
}

function initDrivers() {
    apipolcard.initModule(nativeTest, api_definition);
    apitools.initModule(nativeTest);
}

module.exports.init = init;
